"""Tests for the project settings API in the projectroles app"""

from test_plus.test import TestCase

from ..models import Role, AppSetting, SODAR_CONSTANTS
from ..plugins import get_app_plugin
from ..app_settings import AppSettingAPI
from .test_models import ProjectMixin, RoleAssignmentMixin, AppSettingMixin

# SODAR constants
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
SUBMIT_STATUS_OK = SODAR_CONSTANTS['SUBMIT_STATUS_OK']
SUBMIT_STATUS_PENDING = SODAR_CONSTANTS['SUBMIT_STATUS_PENDING']
SUBMIT_STATUS_PENDING_TASKFLOW = SODAR_CONSTANTS['SUBMIT_STATUS_PENDING']
APP_SETTING_SCOPE_PROJECT = SODAR_CONSTANTS['APP_SETTING_SCOPE_PROJECT']
APP_SETTING_SCOPE_USER = SODAR_CONSTANTS['APP_SETTING_SCOPE_USER']
APP_SETTING_SCOPE_PROJECT_USER = SODAR_CONSTANTS[
    'APP_SETTING_SCOPE_PROJECT_USER'
]

# Local settings
EXISTING_SETTING = 'project_bool_setting'
EXAMPLE_APP_NAME = 'example_project_app'

# App settings API
app_settings = AppSettingAPI()


class TestAppSettingAPI(
    ProjectMixin, RoleAssignmentMixin, AppSettingMixin, TestCase
):
    """Tests for AppSettingAPI"""

    # NOTE: This assumes an example app is available

    def setUp(self):
        self.maxDiff = None

        # Init project
        self.project = self._make_project(
            title='TestProject', type=PROJECT_TYPE_PROJECT, parent=None
        )

        # Init role
        self.role_owner = Role.objects.get(name=PROJECT_ROLE_OWNER)

        # Init user & role
        self.user = self.make_user('owner')
        self.owner_as = self._make_assignment(
            self.project, self.user, self.role_owner
        )

        # Init test project settings
        self.setting_str_values = {
            'app_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'name': 'project_str_setting',
            'setting_type': 'STRING',
            'value': 'test',
            'update_value': 'better test',
            'non_valid_value': False,
        }
        self.setting_int_values = {
            'app_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'name': 'project_int_setting',
            'setting_type': 'INTEGER',
            'value': 0,
            'update_value': 170,
            'non_valid_value': 'Nan',
        }
        self.setting_str_values_options = {
            'app_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'name': 'project_str_setting_options',
            'setting_type': 'STRING',
            'value': 'string1',
            'options': [
                'string1',
                'string2',
            ],  # Options must match the entry in example_project_app.plugins.ProjectAppPlugin
            'update_value': 'string2',
            'non_valid_value': 'string3',
        }
        self.setting_int_values_options = {
            'app_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'name': 'project_int_setting_options',
            'setting_type': 'INTEGER',
            'value': 0,
            'options': [
                0,
                1,
            ],  # Options must match the entry in example_project_app.plugins.ProjectAppPlugin
            'update_value': 1,
            'non_valid_value': 2,
        }
        self.setting_bool_values = {
            'app_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'name': 'project_bool_setting',
            'setting_type': 'BOOLEAN',
            'value': False,
            'update_value': True,
            'non_valid_value': 170,
        }
        self.setting_json_values = {
            'app_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'name': 'project_json_setting',
            'setting_type': 'JSON',
            'value': {
                'Example': 'Value',
                'list': [1, 2, 3, 4, 5],
                'level_6': False,
            },
            'update_value': {'Test_more': 'often_always'},
            'non_valid_value': self.project,
        }

        # Init test user settings
        self.user_setting_str_values = {
            'app_name': EXAMPLE_APP_NAME,
            'user': self.user,
            'name': 'user_str_setting',
            'setting_type': 'STRING',
            'value': 'test',
            'update_value': 'better test',
            'non_valid_value': False,
        }
        self.user_setting_int_values = {
            'app_name': EXAMPLE_APP_NAME,
            'user': self.user,
            'name': 'user_int_setting',
            'setting_type': 'INTEGER',
            'value': 0,
            'update_value': 170,
            'non_valid_value': 'Nan',
        }
        self.user_setting_str_values_options = {
            'app_name': EXAMPLE_APP_NAME,
            'user': self.user,
            'name': 'user_str_setting_options',
            'setting_type': 'STRING',
            'value': 'string1',
            'update_value': 'string2',
            'options': [
                'string1',
                'string2',
            ],  # Options must match the entry in example_project_app.plugins.ProjectAppPlugin
            'non_valid_value': False,
        }
        self.user_setting_int_values_options = {
            'app_name': EXAMPLE_APP_NAME,
            'user': self.user,
            'name': 'user_int_setting_options',
            'setting_type': 'INTEGER',
            'value': 0,
            'update_value': 1,
            'options': [
                0,
                1,
            ],  # Options must match the entry in example_project_app.plugins.ProjectAppPlugin
            'non_valid_value': 'Nan',
        }
        self.user_setting_bool_values = {
            'app_name': EXAMPLE_APP_NAME,
            'user': self.user,
            'name': 'user_bool_setting',
            'setting_type': 'BOOLEAN',
            'value': False,
            'update_value': True,
            'non_valid_value': 170,
        }
        self.user_setting_json_values = {
            'app_name': EXAMPLE_APP_NAME,
            'user': self.user,
            'name': 'user_json_setting',
            'setting_type': 'JSON',
            'value': {
                'Example': 'Value',
                'list': [1, 2, 3, 4, 5],
                'level_6': False,
            },
            'update_value': {'Test_more': 'often_always'},
            'non_valid_value': self.project,
        }

        # Init test project-user settings
        self.project_user_setting_str_values = {
            'app_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'user': self.user,
            'name': 'project_user_string_setting',
            'setting_type': 'STRING',
            'value': 'test',
            'update_value': 'better test',
            'non_valid_value': False,
        }
        self.project_user_setting_int_values = {
            'app_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'user': self.user,
            'name': 'project_user_int_setting',
            'setting_type': 'INTEGER',
            'value': 0,
            'update_value': 170,
            'non_valid_value': 'Nan',
        }
        self.project_user_setting_bool_values = {
            'app_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'user': self.user,
            'name': 'project_user_bool_setting',
            'setting_type': 'BOOLEAN',
            'value': False,
            'update_value': True,
            'non_valid_value': 170,
        }
        self.project_user_setting_json_values = {
            'app_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'user': self.user,
            'name': 'project_user_json_setting',
            'setting_type': 'JSON',
            'value': {
                'Example': 'Value',
                'list': [1, 2, 3, 4, 5],
                'level_6': False,
            },
            'update_value': {'Test_more': 'often_always'},
            'non_valid_value': self.project,
        }

        self.settings = [
            self.setting_int_values,
            self.setting_int_values_options,
            self.setting_json_values,
            self.setting_str_values,
            self.setting_str_values_options,
            self.setting_bool_values,
            self.user_setting_int_values,
            self.user_setting_int_values_options,
            self.user_setting_json_values,
            self.user_setting_str_values,
            self.user_setting_str_values_options,
            self.user_setting_bool_values,
            self.project_user_setting_int_values,
            self.project_user_setting_json_values,
            self.project_user_setting_str_values,
            self.project_user_setting_bool_values,
        ]
        for s in self.settings:
            data = {
                'app_name': s['app_name'],
                'name': s['name'],
                'setting_type': s['setting_type'],
                'value': s['value'] if s['setting_type'] != 'JSON' else '',
                'value_json': s['value'] if s['setting_type'] == 'JSON' else {},
            }
            if 'project' in s:
                data['project'] = s['project']
            if 'user' in s:
                data['user'] = s['user']
            self._make_setting(**data)

    def test_get_project_setting(self):
        """Test get_app_setting()"""
        for setting in self.settings:
            data = {
                'app_name': setting['app_name'],
                'setting_name': setting['name'],
            }
            if 'project' in setting:
                data['project'] = setting['project']
            if 'user' in setting:
                data['user'] = setting['user']
            val = app_settings.get_app_setting(**data)
            self.assertEqual(val, setting['value'])

    def test_get_project_setting_default(self):
        """Test get_app_setting() with default value for existing setting"""
        app_plugin = get_app_plugin(EXAMPLE_APP_NAME)
        default_val = app_plugin.app_settings[EXISTING_SETTING]['default']

        val = app_settings.get_app_setting(
            app_name=EXAMPLE_APP_NAME,
            setting_name=EXISTING_SETTING,
            project=self.project,
        )

        self.assertEqual(val, default_val)

    def test_get_project_setting_nonexisting(self):
        """Test get_app_setting() with an non-existing setting"""
        with self.assertRaises(KeyError):
            app_settings.get_app_setting(
                app_name=EXAMPLE_APP_NAME,
                setting_name='NON-EXISTING SETTING',
                project=self.project,
            )

    def test_get_project_setting_post_safe(self):
        """Test get_app_setting() with JSON setting and post_safe=True"""
        val = app_settings.get_app_setting(
            app_name=self.setting_json_values['app_name'],
            setting_name=self.setting_json_values['name'],
            project=self.setting_json_values['project'],
            post_safe=True,
        )
        self.assertEqual(type(val), str)

    def test_set_project_setting(self):
        """Test set_app_setting()"""

        for setting in self.settings:
            data = {
                'app_name': setting['app_name'],
                'setting_name': setting['name'],
            }
            if 'project' in setting:
                data['project'] = setting['project']
            if 'user' in setting:
                data['user'] = setting['user']

            update_data = dict(data)
            update_data['value'] = setting['update_value']

            ret = app_settings.set_app_setting(**update_data)
            self.assertEqual(ret, True)

            val = app_settings.get_app_setting(**data)
            self.assertEqual(val, setting['update_value'])

    def test_set_project_setting_unchanged(self):
        """Test set_app_setting() with an unchanged value"""

        for setting in self.settings:
            data = {
                'app_name': setting['app_name'],
                'setting_name': setting['name'],
            }
            if 'project' in setting:
                data['project'] = setting['project']
            if 'user' in setting:
                data['user'] = setting['user']

            update_data = dict(data)
            update_data['value'] = setting['value']

            ret = app_settings.set_app_setting(**update_data)
            self.assertEqual(
                ret,
                False,
                msg='setting={}.{}'.format(
                    setting['app_name'], setting['name']
                ),
            )

            val = app_settings.get_app_setting(**data)
            self.assertEqual(
                val,
                setting['value'],
                msg='setting={}.{}'.format(
                    setting['app_name'], setting['name']
                ),
            )

    def test_set_project_setting_new(self):
        """Test set_app_setting() with a new but defined setting"""

        # Assert precondition
        val = AppSetting.objects.get(
            app_plugin=get_app_plugin(EXAMPLE_APP_NAME).get_model(),
            project=self.project,
            name=EXISTING_SETTING,
        ).value
        self.assertEqual(bool(int(val)), False)

        ret = app_settings.set_app_setting(
            app_name=EXAMPLE_APP_NAME,
            setting_name=EXISTING_SETTING,
            value=True,
            project=self.project,
        )

        # Asset postconditions
        self.assertEqual(ret, True)
        val = app_settings.get_app_setting(
            app_name=EXAMPLE_APP_NAME,
            setting_name=EXISTING_SETTING,
            project=self.project,
        )
        self.assertEqual(True, val)

        setting = AppSetting.objects.get(
            app_plugin=get_app_plugin(EXAMPLE_APP_NAME).get_model(),
            project=self.project,
            name=EXISTING_SETTING,
        )
        self.assertIsInstance(setting, AppSetting)

    def test_set_project_setting_undefined(self):
        """Test set_app_setting() with an undefined setting (should fail)"""
        with self.assertRaises(KeyError):
            app_settings.set_app_setting(
                app_name=EXAMPLE_APP_NAME,
                setting_name='new_setting',
                value='new',
                project=self.project,
            )

    def test_set_multi_project_user_settings(self):
        """Test set_app_setting() with multiple instances of a project-user setting"""

        # Set up second user
        new_user = self.make_user('new_user')

        ret = app_settings.set_app_setting(
            app_name=EXAMPLE_APP_NAME,
            setting_name='project_user_string_setting',
            project=self.project,
            user=self.user,
            value=True,
        )
        self.assertEqual(ret, True)

        ret = app_settings.set_app_setting(
            app_name=EXAMPLE_APP_NAME,
            setting_name='project_user_string_setting',
            project=self.project,
            user=new_user,
            value=True,
        )
        self.assertEqual(ret, True)

    def test_validator(self):
        """Test validate_setting() with type BOOLEAN"""
        for setting in self.settings:
            self.assertEqual(
                app_settings.validate_setting(
                    setting['setting_type'],
                    setting['value'],
                    setting.get('options'),
                ),
                True,
            )
            if setting['setting_type'] == 'STRING':
                continue
            with self.assertRaises(ValueError):
                app_settings.validate_setting(
                    setting['setting_type'],
                    setting['non_valid_value'],
                    setting.get('options'),
                )

    def test_validate_project_setting_int(self):
        """Test validate_setting() with type INTEGER"""
        self.assertEqual(
            app_settings.validate_setting('INTEGER', 170, None), True
        )
        # NOTE: String is also OK if it corresponds to an int
        self.assertEqual(
            app_settings.validate_setting('INTEGER', '170', None), True
        )

        with self.assertRaises(ValueError):
            app_settings.validate_setting('INTEGER', 'not an integer', None)

    def test_validate_project_setting_invalid(self):
        """Test validate_setting() with an invalid type"""
        with self.assertRaises(ValueError):
            app_settings.validate_setting('INVALID_TYPE', 'value', None)

    def test_get_setting_def_plugin(self):
        """Test get_setting_def() with a plugin"""
        app_plugin = get_app_plugin(EXAMPLE_APP_NAME)
        expected = {
            'scope': APP_SETTING_SCOPE_PROJECT,
            'type': 'STRING',
            'label': 'String setting',
            'default': '',
            'description': 'Example string project setting',
            'placeholder': 'Example string',
            'user_modifiable': True,
        }
        s_def = app_settings.get_setting_def(
            'project_str_setting', plugin=app_plugin
        )
        self.assertEqual(s_def, expected)

    def test_get_setting_def_app_name(self):
        """Test get_setting_def() with an app name"""
        expected = {
            'scope': APP_SETTING_SCOPE_PROJECT,
            'type': 'STRING',
            'label': 'String setting',
            'default': '',
            'description': 'Example string project setting',
            'placeholder': 'Example string',
            'user_modifiable': True,
        }
        s_def = app_settings.get_setting_def(
            'project_str_setting', app_name=EXAMPLE_APP_NAME
        )
        self.assertEqual(s_def, expected)

    def test_get_setting_def_user(self):
        """Test get_setting_def() with a user setting"""
        expected = {
            'scope': APP_SETTING_SCOPE_USER,
            'type': 'STRING',
            'label': 'String setting',
            'default': '',
            'description': 'Example string user setting',
            'placeholder': 'Example string',
            'user_modifiable': True,
        }
        s_def = app_settings.get_setting_def(
            'user_str_setting', app_name=EXAMPLE_APP_NAME
        )
        self.assertEqual(s_def, expected)

    def test_get_setting_def_invalid(self):
        """Test get_setting_def() with innvalid input"""
        with self.assertRaises(ValueError):
            app_settings.get_setting_def(
                'non_existing_setting', app_name=EXAMPLE_APP_NAME
            )

        with self.assertRaises(ValueError):
            app_settings.get_setting_def(
                'project_str_setting', app_name='non_existing_app_name'
            )

        # Both app_name and plugin unset
        with self.assertRaises(ValueError):
            app_settings.get_setting_def('project_str_setting')

    def test_get_setting_defs_project(self):
        """Test get_setting_defs() with the PROJECT scope"""
        expected = {
            'project_str_setting': {
                'scope': APP_SETTING_SCOPE_PROJECT,
                'type': 'STRING',
                'label': 'String setting',
                'default': '',
                'description': 'Example string project setting',
                'placeholder': 'Example string',
                'user_modifiable': True,
            },
            'project_int_setting': {
                'scope': APP_SETTING_SCOPE_PROJECT,
                'type': 'INTEGER',
                'label': 'Integer setting',
                'default': 0,
                'description': 'Example integer project setting',
                'placeholder': 0,
                'user_modifiable': True,
                'widget_attrs': {'class': 'text-success'},
            },
            'project_str_setting_options': {
                'scope': APP_SETTING_SCOPE_PROJECT,
                'type': 'STRING',
                'label': 'String setting',
                'default': 'string1',
                'options': ['string1', 'string2'],
                'description': 'Example string project setting with options',
                'user_modifiable': True,
            },
            'project_int_setting_options': {
                'scope': APP_SETTING_SCOPE_PROJECT,
                'type': 'INTEGER',
                'label': 'Integer setting',
                'default': 0,
                'options': [0, 1],
                'description': 'Example integer project setting with options',
                'user_modifiable': True,
                'widget_attrs': {'class': 'text-success'},
            },
            'project_bool_setting': {
                'scope': APP_SETTING_SCOPE_PROJECT,
                'type': 'BOOLEAN',
                'label': 'Boolean setting',
                'default': False,
                'description': 'Example boolean project setting',
                'user_modifiable': True,
            },
            'project_global_setting': {
                'scope': SODAR_CONSTANTS['APP_SETTING_SCOPE_PROJECT'],
                'type': 'BOOLEAN',
                'label': 'Global boolean setting',
                'default': False,
                'description': 'Example global boolean project setting',
                'user_modifiable': True,
                'local': False,
            },
            'project_json_setting': {
                'scope': APP_SETTING_SCOPE_PROJECT,
                'type': 'JSON',
                'label': 'JSON setting',
                'default': {
                    'Example': 'Value',
                    'list': [1, 2, 3, 4, 5],
                    'level_6': False,
                },
                'description': 'Example JSON project setting',
                'user_modifiable': True,
                'widget_attrs': {'class': 'text-danger'},
            },
            'project_hidden_setting': {
                'scope': APP_SETTING_SCOPE_PROJECT,
                'type': 'STRING',
                'label': 'Hidden setting',
                'default': '',
                'description': 'Example hidden project setting',
                'user_modifiable': False,
            },
            'project_hidden_json_setting': {
                'scope': APP_SETTING_SCOPE_PROJECT,
                'type': 'JSON',
                'label': 'Hidden JSON setting',
                'description': 'Example hidden JSON project setting',
                'user_modifiable': False,
            },
        }
        defs = app_settings.get_setting_defs(
            APP_SETTING_SCOPE_PROJECT, app_name=EXAMPLE_APP_NAME
        )
        self.assertEqual(defs, expected)

    def test_get_setting_defs_user(self):
        """Test get_setting_defs() with the USER scope"""
        expected = {
            'user_str_setting': {
                'scope': APP_SETTING_SCOPE_USER,
                'type': 'STRING',
                'label': 'String setting',
                'default': '',
                'description': 'Example string user setting',
                'placeholder': 'Example string',
                'user_modifiable': True,
            },
            'user_int_setting': {
                'scope': APP_SETTING_SCOPE_USER,
                'type': 'INTEGER',
                'label': 'Integer setting',
                'default': 0,
                'description': 'Example integer user setting',
                'placeholder': 0,
                'user_modifiable': True,
                'widget_attrs': {'class': 'text-success'},
            },
            'user_str_setting_options': {
                'scope': APP_SETTING_SCOPE_USER,
                'type': 'STRING',
                'label': 'String setting',
                'default': 'string1',
                'options': ['string1', 'string2'],
                'description': 'Example string user setting with options',
                'user_modifiable': True,
            },
            'user_int_setting_options': {
                'scope': APP_SETTING_SCOPE_USER,
                'type': 'INTEGER',
                'label': 'Integer setting',
                'default': 0,
                'options': [0, 1],
                'description': 'Example integer user setting with options',
                'user_modifiable': True,
                'widget_attrs': {'class': 'text-success'},
            },
            'user_bool_setting': {
                'scope': APP_SETTING_SCOPE_USER,
                'type': 'BOOLEAN',
                'label': 'Boolean setting',
                'default': False,
                'description': 'Example boolean user setting',
                'user_modifiable': True,
            },
            'user_json_setting': {
                'scope': APP_SETTING_SCOPE_USER,
                'type': 'JSON',
                'label': 'JSON setting',
                'default': {
                    'Example': 'Value',
                    'list': [1, 2, 3, 4, 5],
                    'level_6': False,
                },
                'description': 'Example JSON user setting',
                'user_modifiable': True,
                'widget_attrs': {'class': 'text-danger'},
            },
            'user_hidden_setting': {
                'scope': APP_SETTING_SCOPE_USER,
                'type': 'STRING',
                'default': '',
                'description': 'Example hidden user setting',
                'user_modifiable': False,
            },
        }
        defs = app_settings.get_setting_defs(
            APP_SETTING_SCOPE_USER, app_name=EXAMPLE_APP_NAME
        )
        self.assertEqual(defs, expected)

    def test_get_setting_defs_project_user(self):
        """Test get_setting_defs() with the PROJECT_USER scope"""
        expected = {
            'project_user_string_setting': {
                'scope': SODAR_CONSTANTS['APP_SETTING_SCOPE_PROJECT_USER'],
                'type': 'STRING',
                'default': '',
                'description': 'Example string project user setting',
            },
            'project_user_int_setting': {
                'scope': SODAR_CONSTANTS['APP_SETTING_SCOPE_PROJECT_USER'],
                'type': 'INTEGER',
                'default': '',
                'description': 'Example int project user setting',
            },
            'project_user_bool_setting': {
                'scope': SODAR_CONSTANTS['APP_SETTING_SCOPE_PROJECT_USER'],
                'type': 'BOOLEAN',
                'default': '',
                'description': 'Example bool project user setting',
            },
            'project_user_json_setting': {
                'scope': SODAR_CONSTANTS['APP_SETTING_SCOPE_PROJECT_USER'],
                'type': 'JSON',
                'default': '',
                'description': 'Example json project user setting',
            },
        }
        defs = app_settings.get_setting_defs(
            APP_SETTING_SCOPE_PROJECT_USER, app_name=EXAMPLE_APP_NAME
        )
        self.assertEqual(defs, expected)

    def test_get_setting_defs_modifiable(self):
        """Test get_setting_defs() with the user_modifiable arg"""
        defs = app_settings.get_setting_defs(
            APP_SETTING_SCOPE_PROJECT, app_name=EXAMPLE_APP_NAME
        )
        self.assertEqual(len(defs), 9)
        defs = app_settings.get_setting_defs(
            APP_SETTING_SCOPE_PROJECT,
            app_name=EXAMPLE_APP_NAME,
            user_modifiable=True,
        )
        self.assertEqual(len(defs), 7)

    def test_get_setting_defs_invalid_scope(self):
        """Test get_setting_defs() with an invalid scope"""
        with self.assertRaises(ValueError):
            app_settings.get_setting_defs(
                'Ri4thai8aez5ooRa', app_name=EXAMPLE_APP_NAME
            )

    def test_get_all_defaults_project(self):
        """Test get_all_defaults() with the PROJECT scope"""
        prefix = 'settings.{}.'.format(EXAMPLE_APP_NAME)
        defaults = app_settings.get_all_defaults(APP_SETTING_SCOPE_PROJECT)
        self.assertEqual(defaults[prefix + 'project_str_setting'], '')
        self.assertEqual(defaults[prefix + 'project_int_setting'], 0)
        self.assertEqual(defaults[prefix + 'project_bool_setting'], False)
        self.assertEqual(
            defaults[prefix + 'project_json_setting'],
            {'Example': 'Value', 'list': [1, 2, 3, 4, 5], 'level_6': False},
        )

    def test_get_all_defaults_user(self):
        """Test get_all_defaults() with the USER scope"""
        prefix = 'settings.{}.'.format(EXAMPLE_APP_NAME)
        defaults = app_settings.get_all_defaults(APP_SETTING_SCOPE_USER)
        self.assertEqual(defaults[prefix + 'user_str_setting'], '')
        self.assertEqual(defaults[prefix + 'user_int_setting'], 0)
        self.assertEqual(defaults[prefix + 'user_bool_setting'], False)
        self.assertEqual(
            defaults[prefix + 'user_json_setting'],
            {'Example': 'Value', 'list': [1, 2, 3, 4, 5], 'level_6': False},
        )

    def test_delete_setting_scope_user_params_none(self):
        self.assertEqual(AppSetting.objects.count(), 16)
        app_settings.delete_setting(EXAMPLE_APP_NAME, 'user_str_setting')
        self.assertEqual(AppSetting.objects.count(), 15)

    def test_delete_setting_scope_user_params_user(self):
        self.assertEqual(AppSetting.objects.count(), 16)
        app_settings.delete_setting(
            EXAMPLE_APP_NAME, 'user_str_setting', user=self.user
        )
        self.assertEqual(AppSetting.objects.count(), 15)

    def test_delete_setting_scope_user_params_project(self):
        with self.assertRaises(ValueError):
            app_settings.delete_setting(
                EXAMPLE_APP_NAME, 'user_str_setting', project=self.project
            )

    def test_delete_setting_scope_user_params_user_project(self):
        with self.assertRaises(ValueError):
            app_settings.delete_setting(
                EXAMPLE_APP_NAME,
                'user_str_setting',
                project=self.project,
                user=self.user,
            )

    def test_delete_setting_scope_project_user_params_none(self):
        self.assertEqual(AppSetting.objects.count(), 16)
        app_settings.delete_setting(
            EXAMPLE_APP_NAME, 'project_user_string_setting'
        )
        self.assertEqual(AppSetting.objects.count(), 15)

    def test_delete_setting_scope_project_user_params_user(self):
        self.assertEqual(AppSetting.objects.count(), 16)
        app_settings.delete_setting(
            EXAMPLE_APP_NAME,
            'project_user_string_setting',
            project=self.project,
            user=self.user,
        )
        self.assertEqual(AppSetting.objects.count(), 15)

    def test_delete_setting_scope_project_user_params_project(self):
        self.assertEqual(AppSetting.objects.count(), 16)
        app_settings.delete_setting(
            EXAMPLE_APP_NAME,
            'project_user_string_setting',
            project=self.project,
            user=self.user,
        )
        self.assertEqual(AppSetting.objects.count(), 15)

    def test_delete_setting_scope_project_user_params_user_project(self):
        self.assertEqual(AppSetting.objects.count(), 16)
        app_settings.delete_setting(
            EXAMPLE_APP_NAME,
            'project_user_string_setting',
            project=self.project,
            user=self.user,
        )
        self.assertEqual(AppSetting.objects.count(), 15)

    def test_delete_setting_scope_project_params_none(self):
        self.assertEqual(AppSetting.objects.count(), 16)
        app_settings.delete_setting(
            EXAMPLE_APP_NAME,
            'project_str_setting',
        )
        self.assertEqual(AppSetting.objects.count(), 15)

    def test_delete_setting_scope_project_params_user(self):
        with self.assertRaises(ValueError):
            app_settings.delete_setting(
                EXAMPLE_APP_NAME,
                'project_str_setting',
                user=self.user,
            )

    def test_delete_setting_scope_project_params_project(self):
        self.assertEqual(AppSetting.objects.count(), 16)
        app_settings.delete_setting(
            EXAMPLE_APP_NAME,
            'project_str_setting',
            project=self.project,
        )
        self.assertEqual(AppSetting.objects.count(), 15)

    def test_delete_setting_scope_project_params_user_project(self):
        with self.assertRaises(ValueError):
            app_settings.delete_setting(
                EXAMPLE_APP_NAME,
                'project_str_setting',
                project=self.project,
                user=self.user,
            )



import re
import sys

from django.contrib import auth
from django.core.management.base import BaseCommand
from django.db import transaction

from projectroles.management.logging import ManagementCommandLogger
from projectroles.models import RemoteSite, SODAR_CONSTANTS


User = auth.get_user_model()
logger = ManagementCommandLogger(__name__)


# SODAR constants
SITE_MODE_TARGET = SODAR_CONSTANTS['SITE_MODE_TARGET']
SITE_MODE_SOURCE = SODAR_CONSTANTS['SITE_MODE_SOURCE']
SITE_MODE_PEER = SODAR_CONSTANTS['SITE_MODE_PEER']


class Command(BaseCommand):
    help = 'Creates a remote site from given arguments'

    def add_arguments(self, parser):
        # ---------------------------
        # RemoteSite Model Argumments
        # ---------------------------
        parser.add_argument(
            '-n',
            '--name',
            dest='name',
            type=str,
            required=True,
            help='Name of the remote site',
        )
        parser.add_argument(
            '-u',
            '--url',
            dest='url',
            type=str,
            required=True,
            help='URL of the remote site. Can be provided without protocol '
            'prefix, defaults to HTTP',
        )
        parser.add_argument(
            '-m',
            '--mode',
            dest='mode',
            type=str,
            required=True,
            help='Mode of the remote site',
        )
        parser.add_argument(
            '-d',
            '--description',
            dest='description',
            type=str,
            required=False,
            default='',
            help='Description of the remote site',
        )
        parser.add_argument(
            '-t',
            '--token',
            dest='secret',
            type=str,
            required=True,
            help='Secret token of the remote site',
        )
        parser.add_argument(
            '-ud',
            '--user-display',
            dest='user_display',
            default=True,
            required=False,
            type=bool,
            help='User display of the remote site',
        )
        # --------------------
        # Additional Arguments
        # --------------------
        parser.add_argument(
            '-s',
            '--suppress-error',
            dest='suppress_error',
            required=False,
            default=False,
            action='store_true',
            help='Suppresses error if site already exists',
        )

    def handle(self, *args, **options):
        logger.info('Creating remote site..')
        name = options['name']
        url = options['url']
        # Validate url
        if not url.startswith('http://') and not url.startswith('https://'):
            url = ''.join(['http://', url])
        pattern = re.compile(r'(http|https)://.*\..*')
        if not pattern.match(url):
            logger.error('Invalid URL "{}"'.format(url))
            sys.exit(1)

        mode = options['mode'].upper()
        # Validate mode
        if mode not in [SITE_MODE_SOURCE, SITE_MODE_TARGET]:
            if mode in [SITE_MODE_PEER]:
                logger.error('Creating PEER sites is not allowed')
            else:
                logger.error('Unkown mode "{}"'.format(mode))
            sys.exit(1)

        description = options['description']
        secret = options['secret']
        user_diplsay = options['user_display']
        suppress_error = options['suppress_error']

        # Validate whether site exists
        name_exists = bool(len(RemoteSite.objects.filter(name=name)))
        url_exists = bool(len(RemoteSite.objects.filter(url=url)))
        if name_exists or url_exists:
            err_msg = 'Remote site exists with {} "{}"'.format(
                'name' if name_exists else 'URL', name if name_exists else url
            )
            if not suppress_error:
                logger.error(err_msg)
                sys.exit(1)
            else:
                logger.info(err_msg)
                sys.exit(0)

        with transaction.atomic():
            create_values = {
                'name': name,
                'url': url,
                'mode': mode,
                'description': description,
                'secret': secret,
                'user_display': user_diplsay,
            }
            site = RemoteSite.objects.create(**create_values)

        logger.info(
            'Created remote site "{}" with mode {}'.format(site.name, site.mode)
        )



"""REST API view tests for the projectroles app with SODAR Taskflow enabled"""

from django.conf import settings
from django.contrib import auth
from django.core.exceptions import ImproperlyConfigured
from django.forms.models import model_to_dict
from django.test import tag
from django.urls import reverse

from rest_framework.test import APILiveServerTestCase

from unittest import skipIf

from projectroles.app_settings import AppSettingAPI
from projectroles.models import Project, Role, RoleAssignment, SODAR_CONSTANTS
from projectroles.plugins import get_backend_api, change_plugin_status
from projectroles.tests.taskflow_testcase import TestCase
from projectroles.tests.test_models import ProjectMixin, RoleAssignmentMixin
from projectroles.tests.test_views_api import SODARAPIViewTestMixin
from projectroles.views_api import CORE_API_MEDIA_TYPE, CORE_API_DEFAULT_VERSION


# SODAR constants
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
SUBMIT_STATUS_OK = SODAR_CONSTANTS['SUBMIT_STATUS_OK']
SUBMIT_STATUS_PENDING = SODAR_CONSTANTS['SUBMIT_STATUS_PENDING']
SUBMIT_STATUS_PENDING_TASKFLOW = SODAR_CONSTANTS[
    'SUBMIT_STATUS_PENDING_TASKFLOW'
]
APP_SETTING_SCOPE_PROJECT = SODAR_CONSTANTS['APP_SETTING_SCOPE_PROJECT']

# Local constants
INVITE_EMAIL = 'test@example.com'
SECRET = 'rsd886hi8276nypuvw066sbvv0rb2a6x'
TASKFLOW_ENABLED = (
    True if 'taskflow' in settings.ENABLED_BACKEND_PLUGINS else False
)
TASKFLOW_SKIP_MSG = 'Taskflow not enabled in settings'
TASKFLOW_TEST_MODE = getattr(settings, 'TASKFLOW_TEST_MODE', False)
NEW_PROJECT_TITLE = 'New Project'
UPDATED_TITLE = 'Updated Title'
UPDATED_DESC = 'Updated description'
UPDATED_README = 'Updated readme'

# Access Django user model
User = auth.get_user_model()

# App settings API
app_settings = AppSettingAPI()


# Tests with Taskflow ----------------------------------------------------------


@tag('Taskflow')  # Run tests if iRODS and SODAR Taskflow are enabled
class TestTaskflowAPIBase(
    ProjectMixin,
    RoleAssignmentMixin,
    SODARAPIViewTestMixin,
    APILiveServerTestCase,
    TestCase,
):
    """Base class for testing API views with taskflow"""

    def _make_project_taskflow(
        self, title, type, parent, owner, description='', readme=''
    ):
        """Make Project with taskflow for API view tests."""
        post_data = dict(self.request_data)
        post_data.update(
            {
                'title': title,
                'type': type,
                'parent': parent.sodar_uuid if parent else None,
                'owner': owner.sodar_uuid,
                'description': description,
                'readme': readme,
            }
        )
        response = self.request_knox(
            reverse('projectroles:api_project_create'),
            method='POST',
            data=post_data,
            media_type=CORE_API_MEDIA_TYPE,
            version=CORE_API_DEFAULT_VERSION,
        )

        # Assert response and object status
        self.assertEqual(response.status_code, 201, msg=response.content)
        project = Project.objects.get(title=title)
        return project, project.get_owner()

    def _make_assignment_taskflow(self, project, user, role):
        """Make RoleAssignment with taskflow for API view tests."""
        url = reverse(
            'projectroles:api_role_create',
            kwargs={'project': project.sodar_uuid},
        )
        self.request_data.update(
            {'role': role.name, 'user': str(user.sodar_uuid)}
        )
        response = self.request_knox(
            url,
            method='POST',
            data=self.request_data,
            media_type=CORE_API_MEDIA_TYPE,
            version=CORE_API_DEFAULT_VERSION,
        )
        self.assertEqual(response.status_code, 201, msg=response.content)
        return RoleAssignment.objects.get(project=project, user=user, role=role)

    def setUp(self):
        # Ensure TASKFLOW_TEST_MODE is True to avoid data loss
        if not TASKFLOW_TEST_MODE:
            raise ImproperlyConfigured(
                'TASKFLOW_TEST_MODE not True, '
                'testing with SODAR Taskflow disabled'
            )

        # Set up live server URL for requests
        self.request_data = {'sodar_url': self.live_server_url}

        # Get taskflow plugin (or None if taskflow not enabled)
        change_plugin_status(
            name='taskflow', status=0, plugin_type='backend'  # 0 = Enabled
        )
        self.taskflow = get_backend_api('taskflow', force=True)

        # Init roles
        self.role_owner = Role.objects.get_or_create(name=PROJECT_ROLE_OWNER)[0]
        self.role_delegate = Role.objects.get_or_create(
            name=PROJECT_ROLE_DELEGATE
        )[0]
        self.role_contributor = Role.objects.get_or_create(
            name=PROJECT_ROLE_CONTRIBUTOR
        )[0]
        self.role_guest = Role.objects.get_or_create(name=PROJECT_ROLE_GUEST)[0]

        # Init user
        self.user = self.make_user('superuser')
        self.user.is_staff = True
        self.user.is_superuser = True
        self.user.save()

        # Create category locally (categories are not handled with taskflow)
        self.category = self._make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self._make_assignment(self.category, self.user, self.role_owner)

        # Get knox token for self.user
        self.knox_token = self.get_token(self.user)

    def tearDown(self):
        self.taskflow.cleanup()


@tag('Taskflow')  # Run tests if iRODS and SODAR Taskflow are enabled
class TestCoreTaskflowAPIBase(TestTaskflowAPIBase):
    """Override of TestTaskflowAPIBase for SODAR Core API views"""

    media_type = CORE_API_MEDIA_TYPE
    api_version = CORE_API_DEFAULT_VERSION


@skipIf(not TASKFLOW_ENABLED, TASKFLOW_SKIP_MSG)
class TestProjectCreateAPIView(TestCoreTaskflowAPIBase):
    """Tests for ProjectCreateAPIView with taskflow"""

    def test_create_project(self):
        """Test project creation"""

        # Assert precondition
        self.assertEqual(Project.objects.all().count(), 1)

        url = reverse('projectroles:api_project_create')
        self.request_data.update(
            {
                'title': NEW_PROJECT_TITLE,
                'type': PROJECT_TYPE_PROJECT,
                'parent': str(self.category.sodar_uuid),
                'description': 'description',
                'readme': 'readme',
                'owner': str(self.user.sodar_uuid),
            }
        )
        response = self.request_knox(url, method='POST', data=self.request_data)

        # Assert response and object status
        self.assertEqual(response.status_code, 201, msg=response.content)
        self.assertEqual(Project.objects.all().count(), 2)


@skipIf(not TASKFLOW_ENABLED, TASKFLOW_SKIP_MSG)
class TestProjectUpdateAPIView(TestCoreTaskflowAPIBase):
    """Tests for ProjectUpdateAPIView with taskflow"""

    def setUp(self):
        super().setUp()

        # Make project with owner in Taskflow and Django
        self.project, self.owner_as = self._make_project_taskflow(
            title='TestProject',
            type=PROJECT_TYPE_PROJECT,
            parent=self.category,
            owner=self.user,
            description='description',
        )

    def test_put_category(self):
        """Test put() for category updating"""
        new_owner = self.make_user('new_owner')

        # Assert preconditions
        self.assertEqual(Project.objects.count(), 2)

        url = reverse(
            'projectroles:api_project_update',
            kwargs={'project': self.category.sodar_uuid},
        )
        self.request_data.update(
            {
                'title': UPDATED_TITLE,
                'type': PROJECT_TYPE_CATEGORY,
                'parent': '',
                'description': UPDATED_DESC,
                'readme': UPDATED_README,
                'owner': str(new_owner.sodar_uuid),
            }
        )
        response = self.request_knox(url, method='PUT', data=self.request_data)

        # Assert response and project status
        self.assertEqual(response.status_code, 200, msg=response.content)
        self.assertEqual(Project.objects.count(), 2)

        # Assert object content
        self.category.refresh_from_db()
        model_dict = model_to_dict(self.category)
        model_dict['readme'] = model_dict['readme'].raw
        expected = {
            'id': self.category.pk,
            'title': UPDATED_TITLE,
            'type': PROJECT_TYPE_CATEGORY,
            'parent': None,
            'description': UPDATED_DESC,
            'readme': UPDATED_README,
            'public_guest_access': False,
            'submit_status': SODAR_CONSTANTS['SUBMIT_STATUS_OK'],
            'full_title': UPDATED_TITLE,
            'has_public_children': False,
            'sodar_uuid': self.category.sodar_uuid,
        }
        self.assertEqual(model_dict, expected)

        # Assert role assignment
        self.assertEqual(
            RoleAssignment.objects.filter(project=self.category).count(), 1
        )
        self.assertEqual(self.category.get_owner().user, new_owner)

    def test_put_project(self):
        """Test put() for project updating"""
        new_owner = self.make_user('new_owner')

        # Assert preconditions
        self.assertEqual(Project.objects.count(), 2)

        url = reverse(
            'projectroles:api_project_update',
            kwargs={'project': self.project.sodar_uuid},
        )
        self.request_data.update(
            {
                'title': UPDATED_TITLE,
                'type': PROJECT_TYPE_PROJECT,
                'parent': str(self.category.sodar_uuid),
                'description': UPDATED_DESC,
                'readme': UPDATED_README,
                'public_guest_access': True,
                'owner': str(new_owner.sodar_uuid),
            }
        )
        response = self.request_knox(url, method='PUT', data=self.request_data)

        # Assert response and project status
        self.assertEqual(response.status_code, 200, msg=response.content)
        self.assertEqual(Project.objects.count(), 2)

        # Assert object content
        self.project.refresh_from_db()
        model_dict = model_to_dict(self.project)
        model_dict['readme'] = model_dict['readme'].raw
        expected = {
            'id': self.project.pk,
            'title': UPDATED_TITLE,
            'type': PROJECT_TYPE_PROJECT,
            'parent': self.category.pk,
            'description': UPDATED_DESC,
            'readme': UPDATED_README,
            'public_guest_access': True,
            'submit_status': SODAR_CONSTANTS['SUBMIT_STATUS_OK'],
            'full_title': self.category.title + ' / ' + UPDATED_TITLE,
            'has_public_children': False,
            'sodar_uuid': self.project.sodar_uuid,
        }
        self.assertEqual(model_dict, expected)

        # Assert role assignment
        self.assertEqual(
            RoleAssignment.objects.filter(project=self.project).count(), 1
        )
        self.assertEqual(self.project.get_owner().user, new_owner)

    def test_patch_category(self):
        """Test patch() for updating category metadata"""

        # Assert preconditions
        self.assertEqual(Project.objects.count(), 2)

        url = reverse(
            'projectroles:api_project_update',
            kwargs={'project': self.category.sodar_uuid},
        )
        self.request_data.update(
            {
                'title': UPDATED_TITLE,
                'description': UPDATED_DESC,
                'readme': UPDATED_README,
            }
        )
        response = self.request_knox(
            url, method='PATCH', data=self.request_data
        )

        # Assert response and project status
        self.assertEqual(response.status_code, 200, msg=response.content)
        self.assertEqual(Project.objects.count(), 2)

        # Assert object content
        self.category.refresh_from_db()
        model_dict = model_to_dict(self.category)
        model_dict['readme'] = model_dict['readme'].raw
        expected = {
            'id': self.category.pk,
            'title': UPDATED_TITLE,
            'type': PROJECT_TYPE_CATEGORY,
            'parent': None,
            'description': UPDATED_DESC,
            'readme': UPDATED_README,
            'public_guest_access': False,
            'submit_status': SODAR_CONSTANTS['SUBMIT_STATUS_OK'],
            'full_title': UPDATED_TITLE,
            'has_public_children': False,
            'sodar_uuid': self.category.sodar_uuid,
        }
        self.assertEqual(model_dict, expected)

        # Assert role assignment
        self.assertEqual(self.category.get_owner().user, self.user)

    def test_patch_project(self):
        """Test patch() for updating project metadata"""

        # Assert preconditions
        self.assertEqual(Project.objects.count(), 2)

        url = reverse(
            'projectroles:api_project_update',
            kwargs={'project': self.project.sodar_uuid},
        )
        self.request_data.update(
            {
                'title': UPDATED_TITLE,
                'description': UPDATED_DESC,
                'readme': UPDATED_README,
            }
        )
        response = self.request_knox(
            url, method='PATCH', data=self.request_data
        )

        # Assert response and project status
        self.assertEqual(response.status_code, 200, msg=response.content)
        self.assertEqual(Project.objects.count(), 2)

        # Assert object content
        self.project.refresh_from_db()
        model_dict = model_to_dict(self.project)
        model_dict['readme'] = model_dict['readme'].raw
        expected = {
            'id': self.project.pk,
            'title': UPDATED_TITLE,
            'type': PROJECT_TYPE_PROJECT,
            'parent': self.category.pk,
            'description': UPDATED_DESC,
            'readme': UPDATED_README,
            'public_guest_access': False,
            'submit_status': SODAR_CONSTANTS['SUBMIT_STATUS_OK'],
            'full_title': self.category.title + ' / ' + UPDATED_TITLE,
            'has_public_children': False,
            'sodar_uuid': self.project.sodar_uuid,
        }
        self.assertEqual(model_dict, expected)

        # Assert role assignment
        self.assertEqual(self.project.get_owner().user, self.user)

    def test_patch_project_move(self):
        """Test patch() for moving project under a different category"""

        new_category = self._make_project(
            'NewCategory', PROJECT_TYPE_CATEGORY, None
        )
        self._make_assignment(new_category, self.user, self.role_owner)
        url = reverse(
            'projectroles:api_project_update',
            kwargs={'project': self.project.sodar_uuid},
        )
        self.request_data.update({'parent': str(new_category.sodar_uuid)})
        response = self.request_knox(
            url, method='PATCH', data=self.request_data
        )

        # Assert response
        self.assertEqual(response.status_code, 200, msg=response.content)

        # Assert object content
        self.project.refresh_from_db()
        model_dict = model_to_dict(self.project)
        self.assertEqual(model_dict['parent'], new_category.pk)

        # Assert role assignment
        self.assertEqual(self.project.get_owner().user, self.user)


@skipIf(not TASKFLOW_ENABLED, TASKFLOW_SKIP_MSG)
class TestRoleAssignmentCreateAPIView(TestCoreTaskflowAPIBase):
    """Tests for RoleAssignmentCreateAPIView with taskflow"""

    def setUp(self):
        super().setUp()

        # Make project with owner in Taskflow and Django
        self.project, self.owner_as = self._make_project_taskflow(
            title='TestProject',
            type=PROJECT_TYPE_PROJECT,
            parent=self.category,
            owner=self.user,
            description='description',
        )

        # Create user for assignments
        self.assign_user = self.make_user('assign_user')

    def test_create_role(self):
        """Test role assignment creation"""

        # Assert precondition
        self.assertEqual(RoleAssignment.objects.all().count(), 2)

        url = reverse(
            'projectroles:api_role_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        self.request_data.update(
            {
                'role': PROJECT_ROLE_CONTRIBUTOR,
                'user': str(self.assign_user.sodar_uuid),
            }
        )
        response = self.request_knox(url, method='POST', data=self.request_data)

        # Assert response and object status
        self.assertEqual(response.status_code, 201, msg=response.content)
        self.assertEqual(RoleAssignment.objects.all().count(), 3)


@skipIf(not TASKFLOW_ENABLED, TASKFLOW_SKIP_MSG)
class TestRoleAssignmentUpdateAPIView(TestCoreTaskflowAPIBase):
    """Tests for RoleAssignmentUpdateAPIView with taskflow"""

    def setUp(self):
        super().setUp()

        # Make project with owner in Taskflow and Django
        self.project, self.owner_as = self._make_project_taskflow(
            title='TestProject',
            type=PROJECT_TYPE_PROJECT,
            parent=self.category,
            owner=self.user,
            description='description',
        )

        # Create user for assignments
        self.assign_user = self.make_user('assign_user')

        # Make extra assignment with Taskflow
        self.update_as = self._make_assignment_taskflow(
            project=self.project,
            user=self.assign_user,
            role=self.role_contributor,
        )

    def test_put_role(self):
        """Test put() for role assignment updating"""

        # Assert precondition
        self.assertEqual(RoleAssignment.objects.all().count(), 3)

        url = reverse(
            'projectroles:api_role_update',
            kwargs={'roleassignment': self.update_as.sodar_uuid},
        )
        self.request_data.update(
            {
                'role': PROJECT_ROLE_GUEST,
                'user': str(self.assign_user.sodar_uuid),
            }
        )
        response = self.request_knox(url, method='PUT', data=self.request_data)

        # Assert response and object status
        self.assertEqual(response.status_code, 200, msg=response.content)
        self.assertEqual(RoleAssignment.objects.all().count(), 3)

    def test_patch_role(self):
        """Test patch() for role assignment updating"""

        # Assert precondition
        self.assertEqual(RoleAssignment.objects.all().count(), 3)

        url = reverse(
            'projectroles:api_role_update',
            kwargs={'roleassignment': self.update_as.sodar_uuid},
        )
        self.request_data.update({'role': PROJECT_ROLE_GUEST})
        response = self.request_knox(
            url, method='PATCH', data=self.request_data
        )

        # Assert response and object status
        self.assertEqual(response.status_code, 200, msg=response.content)
        self.assertEqual(RoleAssignment.objects.all().count(), 3)


@skipIf(not TASKFLOW_ENABLED, TASKFLOW_SKIP_MSG)
class TestRoleAssignmentDestroyAPIView(TestCoreTaskflowAPIBase):
    """Tests for RoleAssignmentDestroyAPIView with taskflow"""

    def setUp(self):
        super().setUp()

        # Make project with owner in Taskflow and Django
        self.project, self.owner_as = self._make_project_taskflow(
            title='TestProject',
            type=PROJECT_TYPE_PROJECT,
            parent=self.category,
            owner=self.user,
            description='description',
        )

        # Create user for assignments
        self.assign_user = self.make_user('assign_user')

        # Make extra assignment with Taskflow
        self.update_as = self._make_assignment_taskflow(
            project=self.project,
            user=self.assign_user,
            role=self.role_contributor,
        )

    def test_delete_role(self):
        """Test delete() for role assignment deletion"""

        # Assert precondition
        self.assertEqual(RoleAssignment.objects.all().count(), 3)

        url = reverse(
            'projectroles:api_role_destroy',
            kwargs={'roleassignment': self.update_as.sodar_uuid},
        )
        response = self.request_knox(
            url, method='DELETE', data=self.request_data
        )

        # Assert response and object status
        self.assertEqual(response.status_code, 204, msg=response.content)
        self.assertEqual(RoleAssignment.objects.all().count(), 2)


@skipIf(not TASKFLOW_ENABLED, TASKFLOW_SKIP_MSG)
class TestRoleAssignmentOwnerTransferAPIView(TestCoreTaskflowAPIBase):
    """Tests for RoleAssignmentOwnerTransferAPIView"""

    def setUp(self):
        super().setUp()

        # Make project with owner in Taskflow and Django
        self.user_owner = self.make_user('user_owner')
        self.project, self.owner_as = self._make_project_taskflow(
            title='TestProject',
            type=PROJECT_TYPE_PROJECT,
            parent=self.category,
            owner=self.user_owner,
            description='description',
        )

        # Create user for assignments
        self.assign_user = self.make_user('assign_user')

    def test_transfer_owner(self):
        """Test transferring ownership for a project"""

        # Make extra assignment with Taskflow
        self._make_assignment_taskflow(
            project=self.project,
            user=self.assign_user,
            role=self.role_contributor,
        )

        # Assert preconditions
        self.assertEqual(self.project.get_owner().user, self.user_owner)

        url = reverse(
            'projectroles:api_role_owner_transfer',
            kwargs={'project': self.project.sodar_uuid},
        )
        post_data = {
            'new_owner': self.assign_user.username,
            'old_owner_role': self.role_contributor.name,
        }
        response = self.request_knox(url, method='POST', data=post_data)

        # Assert response and project status
        self.assertEqual(response.status_code, 200, msg=response.content)
        self.assertEqual(self.project.get_owner().user, self.assign_user)
        self.assertEqual(
            RoleAssignment.objects.get(
                project=self.project, user=self.user_owner
            ).role,
            self.role_contributor,
        )

    def test_transfer_owner_category(self):
        """Test transferring ownership for a category"""

        # Make extra assignment with Taskflow
        self._make_assignment_taskflow(
            project=self.category,
            user=self.assign_user,
            role=self.role_contributor,
        )
        # Assert preconditions
        self.assertEqual(self.category.get_owner().user, self.user)

        url = reverse(
            'projectroles:api_role_owner_transfer',
            kwargs={'project': self.category.sodar_uuid},
        )
        post_data = {
            'new_owner': self.assign_user.username,
            'old_owner_role': self.role_contributor.name,
        }
        response = self.request_knox(url, method='POST', data=post_data)

        # Assert response and project status
        self.assertEqual(response.status_code, 200, msg=response.content)
        self.assertEqual(self.category.get_owner().user, self.assign_user)

    def test_transfer_owner_inherit(self):
        """Test transferring ownership to an inherited owner"""

        # Make extra assignment with Taskflow
        self._make_assignment_taskflow(
            project=self.project,
            user=self.assign_user,
            role=self.role_contributor,
        )

        # Assert preconditions
        self.assertEqual(self.project.get_owner().user, self.user_owner)

        url = reverse(
            'projectroles:api_role_owner_transfer',
            kwargs={'project': self.project.sodar_uuid},
        )
        post_data = {
            'new_owner': self.user.username,  # self.user = category owner
            'old_owner_role': self.role_contributor.name,
        }
        response = self.request_knox(url, method='POST', data=post_data)

        # Assert response and project status
        self.assertEqual(response.status_code, 200, msg=response.content)
        self.assertEqual(self.project.get_owner().user, self.user)
        self.assertEqual(
            RoleAssignment.objects.get(
                project=self.project, user=self.user_owner
            ).role,
            self.role_contributor,
        )



"""
SODAR project and role management
"""

from ._version import get_versions

__version__ = get_versions()['version']
del get_versions

default_app_config = (
    'projectroles.apps.ProjectrolesConfig'  # pylint: disable=invalid-name
)



"""Permission tests for the appalerts app"""

from django.urls import reverse


# Projectroles dependency
from projectroles.tests.test_permissions import TestPermissionBase

from appalerts.tests.test_models import AppAlertMixin


class TestAppAlertPermissions(AppAlertMixin, TestPermissionBase):
    """Tests for AppAlert permissions"""

    def setUp(self):
        # Create users
        self.superuser = self.make_user('superuser')
        self.superuser.is_superuser = True
        self.superuser.is_staff = True
        self.superuser.save()
        self.regular_user = self.make_user('regular_user')
        self.no_alert_user = self.make_user('no_alert_user')
        # No user
        self.anonymous = None
        # Create alert
        self.alert = self._make_app_alert(
            user=self.regular_user, url=reverse('home')
        )

    def test_list(self):
        """Test permissions for the alert list view"""
        url = reverse('appalerts:list')
        good_users = [
            self.superuser,
            self.regular_user,
            self.no_alert_user,
        ]
        bad_users = [self.anonymous]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, bad_users, 302)

    def test_redirect(self):
        """Test permissions for the alert list view"""
        url = reverse(
            'appalerts:redirect', kwargs={'appalert': self.alert.sodar_uuid}
        )
        bad_url = reverse('appalerts:list')
        good_users = [self.regular_user]
        bad_users = [self.superuser, self.no_alert_user, self.anonymous]
        self.assert_response(
            url, good_users, 302, redirect_user=reverse('home')
        )
        self.assert_response(url, bad_users, 302, redirect_user=bad_url)

    def test_ajax_status(self):
        """Test permissions for the alert status ajax view"""
        url = reverse('appalerts:ajax_status')
        good_users = [self.superuser, self.regular_user, self.no_alert_user]
        bad_users = [self.anonymous]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, bad_users, 403)

    def test_ajax_dismiss(self):
        """Test permissions for the alert dismiss ajax view"""
        url = reverse(
            'appalerts:ajax_dismiss', kwargs={'appalert': self.alert.sodar_uuid}
        )
        good_users = [self.regular_user]
        bad_users = [self.superuser, self.no_alert_user]
        self.assert_response(url, good_users, 200, method='POST')
        self.assert_response(url, bad_users, 404, method='POST')
        self.assert_response(url, self.anonymous, 403, method='POST')



"""URL configuration for the ``bgjobs`` app.

Note that the job detail view is implemented
"""

from django.conf.urls import url
from . import views

app_name = 'bgjobs'

urlpatterns = [
    # List jobs that the user has access to
    url(
        regex=r'^(?P<project>[0-9a-f-]+)/list$',
        view=views.ProjectBackgroundJobView.as_view(),
        name='list',
    ),
    # Clear jobs in project owned by the current user.
    url(
        regex=r'^(?P<project>[0-9a-f-]+)/clear/own$',
        view=views.BackgroundJobClearOwnView.as_view(),
        name='clear_own',
    ),
    # Clear jobs in project regardless of the user.
    url(
        regex=r'^(?P<project>[0-9a-f-]+)/clear/all$',
        view=views.BackgroundJobClearAllView.as_view(),
        name='clear_all',
    ),
    # List global background jobs.
    url(
        regex=r'^list$',
        view=views.GlobalBackgroundJobView.as_view(),
        name='site_list',
    ),
]



"""Ajax API view tests for the projectroles app"""

from django.forms import model_to_dict
from django.urls import reverse

from projectroles.models import ProjectUserTag, PROJECT_TAG_STARRED
from projectroles.tests.test_models import (
    ProjectMixin,
    RoleAssignmentMixin,
    ProjectUserTagMixin,
)
from projectroles.tests.test_views import TestViewsBase, PROJECT_TYPE_PROJECT


class TestProjectStarringAjaxView(
    ProjectMixin, RoleAssignmentMixin, ProjectUserTagMixin, TestViewsBase
):
    """Tests for the project starring Ajax view"""

    def setUp(self):
        super().setUp()

        self.project = self._make_project(
            'TestProject', PROJECT_TYPE_PROJECT, None
        )
        self.owner_as = self._make_assignment(
            self.project, self.user, self.role_owner
        )

    def test_star_project(self):
        """Test project starring"""

        # Assert precondition
        self.assertEqual(ProjectUserTag.objects.all().count(), 0)

        # Issue request
        with self.login(self.user):
            response = self.client.post(
                reverse(
                    'projectroles:ajax_star',
                    kwargs={'project': self.project.sodar_uuid},
                )
            )

        # Assert ProjectUserTag state after creation
        self.assertEqual(ProjectUserTag.objects.all().count(), 1)

        tag = ProjectUserTag.objects.get(
            project=self.project, user=self.user, name=PROJECT_TAG_STARRED
        )
        self.assertIsNotNone(tag)

        expected = {
            'id': tag.pk,
            'project': self.project.pk,
            'user': self.user.pk,
            'name': PROJECT_TAG_STARRED,
            'sodar_uuid': tag.sodar_uuid,
        }

        self.assertEqual(model_to_dict(tag), expected)

        # Assert redirect
        self.assertEqual(response.status_code, 200)

    def test_unstar_project(self):
        """Test project unstarring"""
        self._make_tag(self.project, self.user, name=PROJECT_TAG_STARRED)

        # Assert precondition
        self.assertEqual(ProjectUserTag.objects.all().count(), 1)

        # Issue request
        with self.login(self.user):
            response = self.client.post(
                reverse(
                    'projectroles:ajax_star',
                    kwargs={'project': self.project.sodar_uuid},
                )
            )

        # Assert ProjectUserTag state after creation
        self.assertEqual(ProjectUserTag.objects.all().count(), 0)

        # Assert status code
        self.assertEqual(response.status_code, 200)
